/* ********************************************************************** *
 * グローバルウィンドウ
 * ********************************************************************** */

interface Window {
  readonly imHichee: IImHichee;
}

interface IImHichee {
  /** エレメントレポジトリ */
  UIElementRepository: IUIElementRepository;

  /** コンテナアクションアイテムレポジトリ */
  UIContainerActionItemRepository: IUIContainerActionItemRepository;

  /** コンテナアクション条件レポジトリ */
  UIContainerActionConditionRepository: IUIContainerActionConditionRepository;

  /** コンポーネントプロパティのカテゴリレポジトリ */
  UIComponentPropertyCategoryRepository: IUIComponentPropertyCategoryRepository;

  /** HTML エレメントビルダ */
  HTMLElementBuilder: {
    /** エレメントからインスタンス作成 */
    fromElement: (tag: HTMLElement) => IHTMLElementBuilder;

    /** タグ名からインスタンス作成 */
    createElement: (tagName: string, clazz: new () => HTMLElement) => IHTMLElementBuilder;
  };

  /** ブラウザヘルパ */
  BrowserHelper: {
    /**
     * ブラウザ判別
     *
     * コンテナが動作しているシステムとブラウザの情報を取得します。\
     * IM-BloomMaker 本体では、ブラウザの判別にこのプロパティを使用しています。
     */
    readonly browser: BrowserInformation;

    /**
     * 次の描画タイミングでコールバックを実行
     *
     * カレントスレッドの制御をいったん解放し、`setTimeout` よりも早いタイミングでコールバックを実行します。
     */
    nextTick: (callback: () => void) => void;

    /**
     * resize イベントを発火
     *
     * 手動で resize イベントを発火することにより、デザイナやエレメントに対してサイズの再計算や再レンダリングを促します。
     */
    dispatchResizeEvent: () => void;

    /**
     * 安全に実行可能な `setTimeout`
     *
     * `window` の取得を防ぐための特殊な `setTimeout` です。\
     * 第1引数に文字列は指定できません。
     */
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    safeSetTimeout: (handler: TimerHandler, timeout?: number, ...arguments: any[]) => number;

    /**
     * 安全に実行可能な `setInterval`
     *
     * `window` の取得を防ぐための特殊な `setInterval` です。\
     * 第1引数に文字列は指定できません。
     */
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    safeSetInterval: (handler: TimerHandler, timeout?: number, ...arguments: any[]) => number;

    /**
     * BASE URL を取得
     *
     * 現在のページの BASE URL を取得します。
     */
    getBaseURL: () => string;

    /**
     * アラートダイアログを表示
     *
     * IM-BloomMaker 本体で用意されている独自のアラートダイアログを表示します。\
     * このダイアログは、ブラウザウィンドウの閉じる操作や、他の JavaScript の動作を妨げません。\
     * そのため、このメソッドを呼び出した場合は、`await` または、待機処理を実装していない場合、後続の処理が続行されます。
     *
     * このメソッドは Promise を返却します。\
     * ダイアログで「OK」ボタンがクリックされた場合、Promise の後続処理が実行されます。
     */
    alert: (message?: CSJSPrimitiveAllType) => Promise<void>;

    /**
     * 確認ダイアログを表示
     *
     * IM-BloomMaker 本体で用意されている独自の確認ダイアログを表示します。\
     * このダイアログは、ブラウザウィンドウの閉じる操作や、他の JavaScript の動作を妨げません。\
     * そのため、このメソッドを呼び出した場合は、`await` または、待機処理を実装していない場合、後続の処理が続行されます。
     *
     * このメソッドは Promise を返却します。\
     * ダイアログで「OK」または「キャンセル」ボタンがクリックされた場合、Promise の後続処理が実行されます。\
     * 「OK」ボタンの場合は `true`、「キャンセル」ボタンの場合は `false` が、非同期で返却されます。
     */
    confirm: (message?: CSJSPrimitiveAllType) => Promise<boolean>;

    /**
     * エラーダイアログを表示
     *
     * IM-BloomMaker 本体で用意されている独自のエラーダイアログを表示します。\
     * このダイアログは、ブラウザウィンドウの閉じる操作や、他の JavaScript の動作を妨げません。\
     * そのため、このメソッドを呼び出した場合は、`await` または、待機処理を実装していない場合、後続の処理が続行されます。
     *
     * このメソッドは Promise を返却します。\
     * ダイアログで「OK」ボタンがクリックされた場合、Promise の後続処理が実行されます。
     */
    showErrorDialog: (message?: CSJSPrimitiveAllType) => Promise<void>;
  };

  /** 連想配列 */
  KeyValueMap: new <K extends string, V>() => IKeyValueMap<K, V>;

  /**
   * 高精度小数
   *
   * @param num 分子
   * @param denom 分母（未指定の場合は 1）
   */
  Fraction: new (num: number | string | IFraction, denom?: number) => IFraction;

  /**
   * 時差を考慮した Date
   */
  DateTime: {
    /**
     * Date から DateTime を作成
     *
     * @param date 時差を考慮していない Date
     * @param offset 時差（ミリ秒単位）
     */
    fromDate: (date: Date, offset?: number) => IDateTime;

    /**
     * 日時を表した文字列から DateTime を作成
     *
     * 文字列中に時差情報が含まれない場合は、アカウントコンテキストの時差が使用されます。
     */
    fromString: (string: string) => IDateTime;

    /**
     * エポックミリ秒から DateTime を作成
     *
     * アカウントコンテキストの時差が使用されます。\
     * 時差を明示的に指定する場合は、fromDate を使用してください。
     */
    fromNumber: (number: number) => IDateTime;
  };

  /** Argument ファクトリ */
  ArgumentFactory: {
    /** ArrayArgument を作成 */
    createAsArray: (container: IUIContainer, raw: IArgument[]) => IArrayArgument;

    /** BooleanArgument を作成 */
    createAsBoolean: (container: IUIContainer, raw: boolean) => IArgument;

    /** DateArgument を作成 */
    createAsDate: (container: IUIContainer, raw: string | number | {date: Date; offset?: number}) => IDateArgument;

    /** DoubleArgument を作成 */
    createAsDouble: (container: IUIContainer, raw: number) => INumberArgument;

    /** 空の ObjectArgument を作成 */
    createAsEmptyObject: (container: IUIContainer) => IObjectArgument;

    /** FractionArgument を作成 */
    createAsFraction: (container: IUIContainer, raw: IFraction) => INumberArgument;

    /** IntegerArgument を作成 */
    createAsInteger: (container: IUIContainer, raw: number) => INumberArgument;

    /** NullArgument を作成 */
    createAsNull: (container: IUIContainer) => IArgument;

    /**
     * ObjectArgument を作成
     *
     * 引数の `raw` には `new window.imHichee.KeyValueMap<string, IArgument>()` で作成したインスタンスを指定してください。
     */
    createAsObject: (container: IUIContainer, raw: IKeyValueMap<string, IArgument>) => IObjectArgument;

    /** ParameterPathArgument を作成 */
    createAsParameterPath: (container: IUIContainer, target: IUIComponent, raw: string) => IParameterPathArgument;

    /** StringArgument を作成 */
    createAsString: (container: IUIContainer, raw: string) => IArgument;
  };

  /** エレメントID採番 */
  UIElementIDResolver: IUIElementIDResolver;

  /** メッセージビルダ */
  MessageBuilder: IMessageBuilder;

  /** UUID */
  UUID: {
    /** UUID作成 */
    generate: () => string;
  };

  /** デザイナコンテンツ情報 */
  content: {
    type: string;
    contentId: string;
    categoryId: string;
    cssFramework: string;
    name: string;
    localizedName: {[key: string]: string};
    description: string;
    localizedDescription: {[key: string]: string};
    sourceTemplateId: string;
    sortNumber: number;
    version: string;
    adminFlag: string;
    plugins: {id: string; [key: string]: CSJSPrimitiveSingleType}[];
    token: string;
  };

  /** デザイナテンプレート情報 */
  template: {
    type: string;
    templateId: string;
    categoryId: string;
    cssFramework: string;
    name: string;
    localizedName: {[key: string]: string};
    description: string;
    localizedDescription: {[key: string]: string};
    sourceContentId: string;
    sortNumber: number;
    adminFlag: string;
    plugins: {id: string; [key: string]: CSJSPrimitiveSingleType}[];
    token: string;
  };

  /** デザイナエレメントセット情報 */
  elementSet: {
    type: string;
    elementSetId: string;
    categoryId: string;
    cssFramework: string;
    name: string;
    localizedName: {[key: string]: string};
    description: string;
    localizedDescription: {[key: string]: string};
    sortNumber: number;
    adminFlag: string;
    plugins: {id: string; [key: string]: CSJSPrimitiveSingleType}[];
    token: string;
  };

  /** 設定 */
  configuration: {
    /** レイアウト用 */
    layout: {
      /** elementCount プロパティの最大値 */
      maxElementCount: number;
      /** rowCount プロパティの最大値 */
      maxRowCount: number;
      /** columnCount プロパティの最大値 */
      maxColumnCount: number;
    };
    /** レンダリング用 */
    render: {
      /** 繰り返し要素レンダリング最大時間（ms） */
      maxRepeatableCreateTime: number;
    };
    /** ロード処理用 */
    loading: {
      /** ロード完了を待機するインターバル時間（ms） */
      intervalTime: number;
      /** ロード完了を待機する試行回数 */
      intervalCountThreshold: number;
    };
  };

  /** サーバから取得したコンテキスト */
  contexts: {
    /** アカウントコンテキスト */
    accountContext: {
      /** 多言語のロケール解決に利用するロケールID */
      localeId: string;
      /** DateTime で利用するタイムゾーン */
      timeZone: {
        /** タイムゾーンID */
        id: string;
        /** UTCからのオフセット（ミリ秒） */
        rawOffset: number;
        /** 夏時間 */
        useDaylightTime: boolean;
      };
    };
  };

  /** サーバから取得したシステム情報 */
  systemInfo: {
    /** システム情報 */
    system: {
      /** ロケールID */
      localeId: string;

      /** システムで利用可能なロケール */
      availableLocales: {
        /** ロケールID */
        localeId: string;
        /** ロケール表示名 */
        displayName: string;
        /** デフォルトエンコーディング */
        defaultEncoding: string;
      }[];
    };

    /** テナント情報 */
    tenant: {
      /** ロケールID */
      localeId: string;
    };

    /** インストール済みモジュール */
    installedModules: {
      /** 設計書出力 */
      im_bloommaker_development: boolean;
    };
  };

  /** 環境情報 */
  env: IEnv;

  /** コンテンツ識別（コンテンツ / テンプレート / エレメントセット） */
  readonly contentDefinitionType: 'content' | 'template' | 'element-set';

  /** ImSecureToken */
  readonly imSecureToken: string;

  /** 変数の「入力」です。実行モードでのみ使用されます。 */
  readonly inputData?: string;

  /** コンテンツのデータです。実行モードでのみ使用されます。 */
  readonly serializedData?: string;
}

/** 環境情報 */
interface IEnv {
  accountContext: {
    applicationLicenses: string[];
    authenticated: boolean;
    calendarId: string;
    dateTimeFormats: {
      IM_DATETIME_FORMAT_DATE_STANDARD: string;
      IM_DATETIME_FORMAT_TIME_TIMESTAMP: string;
      IM_DATETIME_FORMAT_TIME_INPUT: string;
      'format-set-id': string;
      IM_DATETIME_FORMAT_TIME_STANDARD: string;
      IM_DATETIME_FORMAT_DATE_SIMPLE: string;
      IM_DATETIME_FORMAT_DATE_INPUT: string;
      locale: string;
    };
    decimalFormatId: string;
    encoding: string;
    firstDayOfWeek: number;
    homeUrl: string;
    locale: string;
    loginTime: string;
    roleIds: string[];
    signature: string;
    tenantId: string;
    themeId: string;
    timeZone: {
      id: string;
      rawOffset: number;
      useDaylightTime: boolean;
    };
    userCd: string;
    userType: string;
  };
  userContext: {
    allDepartments: string[];
    allPosts: string[];
    companyList: string[];
    currentDepartment: string[];
    departmentByCompany: string[];
    mainDepartment: {
      companyCd: string;
      departmentCd: string;
      departmentFullName: string;
      departmentName: string;
      departmentSearchName: string;
      departmentSetCd: string;
      departmentShortName: string;
    };
    mainPostList: string[];
    postByCompany: object[];
    publicGroupList: string[];
    publicGroupRoleList: string[];
    userCategoryList: string[];
    userProfile: {
      address1: string;
      address2: string;
      address3: string;
      countryCd: string;
      emailAddress1: string;
      emailAddress2: string;
      extensionFaxNumber: string;
      extensionNumber: string;
      faxNumber: string;
      mobileEmailAddress: string;
      mobileNumber: string;
      notes: string;
      sex: string;
      telephoneNumber: string;
      url: string;
      userCd: string;
      userName: string;
      userSearchName: string;
      zipCode: string;
    };
  };
  clientInfo: {
    os: string;
    application: string;
  };
  const: {
    startTime: Date;
    true: boolean;
    false: boolean;
    null: null;
  };
  location: {
    base: string;
    path: string;
  };
}

/* ********************************************************************** *
 * コンポーネント
 * ********************************************************************** */

interface IUIComponent {
  /**
   * コンポーネントクラス
   *
   * コンポーネントを構成するクラスを取得します。
   */
  readonly clazz: UIComponentClass;

  /**
   * コンポーネントの配置制約
   *
   * エレメントのコピー、削除、配置などに関する制約を取得します。
   */
  readonly constraint: IUIComponentConstraint;

  /**
   * コンポーネントのプロパティ
   *
   * このコンポーネントが持つプロパティに設定されている生の値を取得します。
   */
  readonly properties: IUIComponentProperties;

  /**
   * 有効状態
   *
   * コンポーネントのインスタンスが作成されてから `destroy` されていない場合、`true` を返却します。
   * `destroy` メソッドなどによりコンテナ上で使用できない状態の場合、`false` を返却します。
   */
  readonly alived: boolean;

  /**
   * レンダリング状態
   *
   * コンテナ上でレンダリングされている（画面上に表示されている）状態の場合、`true` を返却します。
   * レンダリングされていない場合、`false` を返却します。
   */
  readonly rendered: boolean;

  /**
   * レンダリングで使用するラップされた HTML エレメントビルダ
   *
   * コンポーネントを実際にレンダリングする際にルートとなるタグを構成するためのエレメントビルダを取得します。\
   * レンダリング時に、このエレメントビルダを使用して、実態ノードのビルドが行われます。
   */
  readonly wrappedBuilder: IHTMLElementBuilder;

  /**
   * 子エレメントの配列
   *
   * このコンポーネント配下にある子コンポーネントを取得します。\
   * このプロパティが返却するコンポーネントの順序は保証されていません。\
   * コンテナ上で描画される順序で取得する場合、`IUIElementController#getSortedChildren` メソッドを使用してください。
   */
  readonly children: IUIElement[];

  /** コンポーネントごとに割り当てられた UUID */
  readonly uuid: string;

  /** コンポーネントの表示位置 */
  getClientRect(): UIComponentClientRect;

  /** コンポーネントの内部位置 */
  getOffsetRect(): UIComponentClientRect;
}

type UIComponentClass =
  /** コンテナ */
  | 'UIContainer'
  /** エレメント */
  | 'UIElement';

type UIComponentClientRect = {
  left: number; // 一時変数扱いのため、書き換え可能
  right: number; // 一時変数扱いのため、書き換え可能
  top: number; // 一時変数扱いのため、書き換え可能
  bottom: number; // 一時変数扱いのため、書き換え可能
  width: number; // 一時変数扱いのため、書き換え可能
  height: number; // 一時変数扱いのため、書き換え可能
};

interface IUIComponentConstraint {
  /**
   * 移動可否
   *
   * デザイナ上でコンポーネントを移動できる場合、`true` を返却します。\
   * 移動できない場合、`false` を返却します。\
   * この返却値にかかわらず、API 経由で操作した場合は、コンポーネントを移動できます。
   */
  readonly movable: boolean;

  /**
   * 削除可否
   *
   * デザイナ上でコンポーネントを削除できる場合、`true` を返却します。\
   * 削除できない場合、`false` を返却します。\
   * この返却値にかかわらず、API 経由で操作した場合は、コンポーネントを削除できます。
   */
  readonly removable: boolean;

  /**
   * コピー可否
   *
   * デザイナ上でコンポーネントを複製できる場合、`true` を返却します。\
   * 複製できない場合、`false` を返却します。\
   * この返却値にかかわらず、API 経由で操作した場合は、コンポーネントを複製できます。
   */
  readonly copyable: boolean;

  /**
   * 末端フラグ
   *
   * このエレメントより配下にエレメントを配置できない場合、`true` を返却します。\
   * 子エレメントを配置できる場合、`false` を返却します。
   *
   * これは、`acceptableChild` メソッドが常に `false` を返却する場合、`true` を返却し、
   * それ以外であれば `false` を返却しなくてはならないことを意味します。
   */
  readonly extremity: boolean;

  /**
   * 子として、親の受け入れ可否を取得
   *
   * 引数で渡されたエレメントが、親として受け入れられる場合、`true` を返却します。\
   * 受け入れられない場合、`false` を返却します。
   */
  acceptableParent(parent: ParentUIElement): boolean;

  /**
   * 親として、子の受け入れ可否を取得
   *
   * 引数で渡されたエレメントが、子として受け入れられる場合、`true` を返却します。\
   * 受け入れられない場合、`false` を返却します。
   */
  acceptableChild(child: IUIElement): boolean;

  /**
   * 移動元と移動先を考慮して、受け入れ可否を取得
   *
   * 引数で渡された移動元の親エレメントから、移動先の親エレメントへ移動できる場合、`true` を返却します。\
   * 移動できない場合、`false` を返却します。\
   * movable プロパティとこのメソッドの返却値が、両方とも `true` のとき、エレメントは移動可能です。
   *
   * このメソッドを実装しない場合は、movable プロパティの結果のみで判定します。\
   */
  acceptableMove?(before: ParentUIElement, after: ParentUIElement): boolean;
}

interface IUIComponentProperties {
  /** プロパティを取得 */
  getProperty(key: string): IUIComponentProperty;

  /** プロパティを列挙 */
  forEach(callback: UIComponentPropertiesCallback): void;

  /** キーの存在をチェック */
  containsKey(key: string): boolean;
}

type UIComponentPropertiesCallback = (key: string, property: IUIComponentProperty) => void;

interface IUIComponentProperty {
  /** プロパティキー */
  readonly key: string;

  /** 表示名 */
  readonly displayName: string;

  /** プロパティ定義 */
  readonly definition: IUIComponentPropertyDefinition;

  /** 入力タイプ */
  readonly type: UIComponentPropertyType;

  /** プロパティ値 */
  readonly value: IPropertyValueArgument;
}

type UIComponentPropertyType =
  /** 文字列入力 */
  | 'string'
  /** 整数入力 */
  | 'integer'
  /** 数値入力 */
  | 'decimal'
  /** 真偽値入力 */
  | 'boolean'
  /** 入力欄なし */
  | 'none'
  /** アクション選択 */
  | 'action'
  /** ページ選択 */
  | 'page';

/**
 * コンポーネントのプロパティの説明は、`SIT.Z.IWP.HICHEE.COMPONENT.PROPERTY.DESCRIPTION.${コンポーネント名}.${プロパティのキー}` のメッセージプロパティキーで指定します。
 * メッセージプロパティキーが存在しない場合は、プロパティのヘルプアイコンを表示しません。
 */
interface IUIComponentPropertyDefinition {
  /** 必須項目 */
  readonly required: boolean;

  /** データの種別 */
  readonly dataType: UIComponentPropertyDataType;

  /** 許容される最小値 */
  readonly min: number | null;

  /** 許容される最大値 */
  readonly max: number | null;

  /** 読み取り専用（ユーザによる書き換え不可） */
  readonly readonly: boolean;

  /** 値変更時、再レンダリング */
  readonly rerender: boolean;

  /** エレメント繰り返し使用 */
  readonly repeatable: boolean;

  /** プレビュー時および実行画面でのバリデーション実行 */
  readonly validateValue: boolean;

  /** 絶対値指定可能 */
  readonly acceptStatic: boolean;

  /** 相対値指定可能 */
  readonly acceptDynamic: boolean;

  /** 複数行文字列が入力可能 */
  readonly acceptMultipleLines: boolean;

  /** 入力候補 */
  readonly candidate: IUIComponentPropertyCandidate[] | null;
}

type UIComponentPropertyDataType =
  /** 固定小数点型 */
  | 'NUMERIC'
  /** 整数型 */
  | 'INTEGER'
  /** 指定なし */
  | 'ANY';

interface IUIComponentPropertyCandidate {
  /** 表示名 */
  readonly displayName: string;

  /** 実際の値 */
  readonly value: UIComponentPropertyValueType;
}

type UIComponentPropertyValueType = string | number | boolean | null;

interface IUIComponentPropertyCategoryRepository {
  /** プロパティカテゴリの一覧 */
  readonly categories: UIComponentPropertyCategoryInfo[];

  /** プロパティカテゴリの登録 */
  registerCategory(categoryInfo: UIComponentPropertyCategoryInfo): void;
}

/**
 * プロパティのカテゴリ名は、`CAP.Z.IWP.HICHEE.COMPONENT.PROPERTY.CATEGORY.NAME.{カテゴリID}` のメッセージプロパティキーで指定します。
 * このキーが存在しない場合はカテゴリIDを使用します。
 */
type UIComponentPropertyCategoryInfo = {
  /** カテゴリID */
  readonly id: string;

  /** ソート番号（昇順） */
  readonly sortNumber?: number;

  /** 非表示条件とする表示タイプの一覧 */
  readonly hiddenDisplayTypes?: IUIComponentDisplayType[];
};

interface IUIComponentDisplayType {
  /** クラス名 */
  readonly className: string;

  /** 幅自動調整フラグ */
  readonly autoWidth: boolean;

  /** 高さ自動調整フラグ */
  readonly autoHeight: boolean;
}

type ParentUIElement = IUIContainer | IUIElement;
type NextUIElement = IUIElement | null;

type UIComponentConstraintClass = new () => IUIComponentConstraint;

/* ********************************************************************** *
 * コンテナ
 * ********************************************************************** */

interface IUIContainer extends IUIComponent {
  /**
   * デザイナモード
   *
   * デザイナ画面上で使用するコンテナの場合、`true` を返却します。\
   * プレビュー画面、実行画面上で使用するコンテナの場合、`false` を返却します。
   */
  readonly designMode: boolean;

  /**
   * 作業用モード
   *
   * クリップボード目的など、表示以外の目的で使用するコンテナの場合、`true` を返却します。\
   * 画面上に表示する目的で使用するコンテナの場合、`false` を返却します。
   */
  readonly workingMode: boolean;

  /**
   * エレメントレポジトリ
   *
   * このコンテナで使用可能なエレメントを管理するレポジトリを取得します。
   */
  readonly repository: IUIElementRepository;

  /**
   * エレメントコントローラ
   *
   * このコンテナに追加されたエレメントのインスタンスを管理するコントローラを取得します。
   */
  readonly controller: IUIElementController;

  /**
   * パラメータストレージ
   *
   * このコンテナで有効な変数・定数・入力・多言語情報を管理するストレージを取得します。
   */
  readonly parameters: IParameterStorage;

  /**
   * アクションチャートストレージ
   *
   * このコンテナで実行可能なアクションアイテムをまとめたアクションチャートを管理するストレージを取得します。
   */
  readonly eventCharts: IActionChartStorage;

  /**
   * コンテナ
   *
   * 自身を取得します。
   */
  readonly container: IUIContainer;

  /**
   * コンポーネントのプロパティ
   *
   * このコンテナが持つプロパティに設定されている生の値を取得します。
   */
  readonly properties: IUIContainerProperties;

  /**
   * スクロール位置
   *
   * コンテナのサイズ制限をしている場合、コンテナ内部エリアのスクロール位置を取得します。
   */
  readonly scroll: {left: number; top: number};

  /**
   * バリデーションエラー存在フラグ
   *
   * コンテナ内の全てのエレメントで、バリデーションエラーが存在するかどうかを取得します。\
   * ひとつでもバリデーションエラーがある場合は、`true` を返却します。
   */
  readonly hasValidationError: boolean;

  /**
   * レンダリング
   *
   * コンテナの各イベントドライバを初期化し、紐づいているエレメントをレンダリングします。\
   * レンダリング後にもう一度呼び出すと、再レンダリングを行います。
   */
  render(force?: boolean): void;

  /**
   * 削除
   *
   * エレメントを削除すると、自動的に子エレメントも削除され、親子関係は解消します。\
   * コンポーネントを削除すると、復活はできません。
   */
  destroy(): void;

  /**
   * ローディング表示
   *
   * 引数に指定した非同期処理を実行中、コンテナにローディングレイヤを表示します。\
   * 非同期処理が終了すると、自動的にローディングレイヤを削除します。
   */
  showLoadingIndicator<T>(promise: Promise<T>): Promise<T>;

  /**
   * ローディング表示方式を変更
   *
   * 現在表示中のローディングレイヤの表示方法を変更します。\
   * ここで設定した表示方法は、ローディングレイヤが削除されるまで有効です。\
   * ローディングレイヤが削除された場合、次回のローディングレイヤ表示時には、初期状態に戻ります。
   *
   * `showLoadingIndicator` メソッドによってローディングが表示されている場合のみ有効です。\
   * ローディングが表示していないときにメソッドを呼び出しても、何も起こりません。
   */
  changeLoadingIndicatorBehavior(behavior: LoadingIndicatorBehavior): void;
}

interface IUIContainerProperties extends IUIComponentProperties {}

type LoadingIndicatorBehavior = {
  /** アニメーションを中止して、即時表示状態にする */
  cancelAnimation?: true;
  /** 中央のテキストを非表示にする */
  hideIndicatorText?: true;
};

/* ********************************************************************** *
 * エレメント
 * ********************************************************************** */

interface IUIElement extends IUIComponent {
  /**
   * 表示タイプ名
   *
   * ブラウザ画面上にどのように表示するかを示す文字列を取得します。
   */
  readonly displayType: IUIComponentDisplayType;

  /**
   * コンテナ
   *
   * このエレメントが所属するコンテナを取得します。
   */
  readonly container: IUIContainer;

  /**
   * コンポーネントの配置制約
   *
   * エレメントのコピー、削除、配置などに関する制約を取得します。
   */
  readonly constraint: IUIComponentConstraint;

  /**
   * コンポーネントのプロパティ
   *
   * このエレメントが持つプロパティに設定されている生の値を取得します。\
   * 実行画面で固定値・変数値を考慮した最終的な解決済みの値を取得する場合は、`propertyAccessor` を使用してください。
   */
  readonly properties: IUIElementProperties;

  /**
   * プロパティのアクセサ
   *
   * コンポーネントのプロパティに設定されている Argument に対して、値を取得・新しい値を設定する機能を提供します。\
   * 値を取得した場合、固定値・変数値を考慮した最終的な解決済みの値を返却します。
   */
  readonly propertyAccessor: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>;

  /**
   * エレメントのインスタンス
   *
   * エレメントの動作や処理が定義された実態クラスのインスタンスを取得します。
   */
  readonly instance: IUIElementCore;

  /**
   * 元となる HTML エレメントビルダ
   *
   * エレメントのルートとなるタグを構成するためのエレメントビルダを取得します。\
   * レンダリング時に、このエレメントビルダを使用して、実態ノードのビルドが行われます。
   */
  readonly builder: IHTMLElementBuilder;

  /**
   * 親のエレメント
   *
   * このエレメントが配置されている親のエレメント、または、コンテナを取得します。\
   * このプロパティを書き換えると、代入したエレメント、または、コンテナを親として、配下の最後尾へエレメントを移動します。
   */
  parent: ParentUIElement;

  /**
   * 次のエレメント
   *
   * このエレメントの次に配置されているエレメントを取得します。\
   * 自身が最後のエレメントである場合、`null` を返却します。
   *
   * このプロパティを書き換えると、代入したエレメントの手前へエレメントを移動します。\
   * null を代入した場合は、同じ親の最後尾へエレメントを移動します。
   */
  next: NextUIElement;

  /**
   * レンダリング済みの次のエレメント
   *
   * このエレメントの次に配置されている、すでにレンダリング済みのエレメントを取得します。\
   * 自身が最後のエレメントであるか、レンダリング済みエレメントが次にない場合、`null` を返却します。
   */
  readonly renderedNext: NextUIElement;

  /**
   * 表示状態
   *
   * 画面上にノードを配置するかどうかを取得・設定します。\
   * ノードツリー上に配置して画面上にエレメントを表示する場合、`true` を設定します。\
   * ノードツリー上から削除して画面上からエレメントを非表示にする場合、`false` を設定します。
   *
   * 初期状態は `true` です。
   */
  visibility: boolean;

  /**
   * レンダリング
   *
   * 削除状態の場合は何も行いません。\
   * 削除状態を解除して再レンダリングを行う場合は、`resume` メソッドを使用してください。
   */
  render(force?: boolean): void;

  /**
   * 削除
   *
   * エレメントを削除すると、自動的に子エレメントも削除され、親子関係は解消します。\
   * `resume` メソッドを使用しても、子エレメントは復活しません。
   */
  destroy(): void;

  /**
   * 削除を復活
   *
   * `destroy` メソッドによって削除された状態を解除し、インスタンスを復活します。\
   * ただし、子エレメントは復活しません。
   */
  resume(): void;

  /**
   * インスタンスを複製
   *
   * 指定した parent にエレメントのインスタンスを複製して、新しいエレメントを返却します。\
   * 子エレメントも自動的に複製します。
   */
  clone(parent?: ParentUIElement): IUIElement;

  /**
   * プロパティの変数値バリデーションを実行
   *
   * エレメントの各プロパティに指定された変数値の入力規則と、プロパティの入力値を比較して、判定結果を返却します。\
   * ひとつでも入力規則に違反しているプロパティがある場合は、`true` を返却して、実行画面上に赤枠を表示します。\
   * 常に非同期で実行しますので、`await` または、待機処理を適切に実装してください。
   */
  validate(): Promise<boolean>;
}

interface ISimpleUIElement extends IUIElement {}

interface IIndexableUIElement extends IUIElement {
  /**
   * 繰り返しインデックス
   *
   * 親エレメントから見て、自身のエレメントが何番目に位置するかを取得・設定します。\
   * このインデックスが、変数パスの `$index` で使用されます。
   *
   * この値は、エレメントのレンダリング中に IM-BloomMaker 本体によって書き換えられます。\
   * エレメント側からは、値を書き換えないでください。
   */
  repeatIndex: number;
}

interface IRepeatableUIElement extends IIndexableUIElement {}

interface IUIElementCore {
  /**
   * エレメントタイプ名
   *
   * エレメントのクラス名と同一の文字列を取得します。
   */
  readonly elementTypeName: string;

  /**
   * 表示タイプ名
   *
   * ブラウザ画面上にどのように表示するかを示す文字列を取得します。
   */
  readonly displayType: UIElementDisplayType;

  /**
   * エレメントの制約クラス
   *
   * エレメントのコピー、削除、配置などに関する制約をまとめたクラスを取得します。
   */
  readonly constraintClass: UIComponentConstraintClass;

  /**
   * エレメント固有のプロパティ定義
   *
   * エレメントが持つ固有のプロパティ定義を取得します。\
   * キーは自由に設定可能ですが、`commonPropertyDefinition` と重複するキーは指定できません。
   */
  readonly uniquePropertyDefinition: IUniquePropertyDefinition;

  /**
   * エレメント共通のプロパティ定義
   *
   * 全てのエレメントが持つ共通のプロパティ定義を取得します。
   */
  readonly commonPropertyDefinition: ICommonPropertyDefinition;

  /**
   * 依存するエレメントクラス
   *
   * `createChildren`, `updateChildren`, `cloneChildren` で、子エレメントとして自身以外のエレメントを使用する場合、
   * エレメントのクラスを返却すると、IM-Hichee 側によって自動的にクラスが登録され、使用できるようになります。
   */
  readonly dependElements: UIElementCoreClass[];

  /**
   * エレメントラッパークラス
   *
   * 繰り返さないシンプルなエレメントを作成する場合は `SimpleUIElement`、
   * 手動で繰り返すエレメントを作成する場合は `IndexableUIElement`、
   * 自動的に繰り返すエレメントを作成する場合は `RepeatableUIElement` を返却してください。
   */
  readonly wrapperClass: UIElementWrapperClass;

  /**
   * HTML エレメントを作成
   *
   * 仮想エレメントが作成された際に IM-Hichee 側から1度だけ呼び出されるメソッドです。\
   * ルートとなるタグの `IHTMLElementBuilder` インスタンスを作成して、返却してください。\
   * `IHTMLElementBuilder` のインスタンスは、`window.imHichee.HTMLElementBuilder.createElement`
   * または `window.imHichee.HTMLElementBuilder.fromElement` メソッドで作成できます。
   *
   * このメソッドで返却する `IHTMLElementBuilder` のインスタンスでは、
   * `setCSS`, `removeCSS`, `switchCSS` メソッドを使用しないでください。\
   * なお、子エレメントのインスタンスに対しては、制限はありません。
   *
   * `IHTMLElementBuilder` の `build` メソッドは自動で呼び出されますので、明示的に呼び出す必要はありません。
   */
  createElement(
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IHTMLElementBuilder;

  /**
   * HTML エレメントを更新
   *
   * レンダリングが必要になった際に、都度 IM-Hichee 側から呼び出されるメソッドです。\
   * 引数で渡された `IHTMLElementBuilder` のインスタンスで、必要なレンダリング処理を行ってください。
   *
   * `IHTMLElementBuilder` の `build` メソッドは自動で呼び出されますので、明示的に呼び出す必要はありません。
   */
  updateElement(
    builder: IHTMLElementBuilder,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): void;

  /**
   * HTML エレメントを削除
   *
   * 仮想エレメントが削除された際に IM-Hichee 側から1度だけ呼び出されるメソッドです。\
   * 引数で渡された `IHTMLElementBuilder` のインスタンスで、必要な解放処理を行ってください。
   *
   * `IHTMLElementBuilder` に登録された `tag` は自動で親から外されますので、明示的にタグを解除する必要はありません。
   */
  destroyElement(
    builder: IHTMLElementBuilder,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): void;

  /**
   * 子エレメントを作成
   *
   * 仮想エレメントが純粋に（クローンせずに）作成された際に IM-Hichee 側から1度だけ呼び出されるメソッドです。\
   * このエレメントが作成された際に自動的に用意する子エレメントのインスタンスを返却してください。
   *
   * 子エレメントは配列の格納順に配置されます。
   */
  createChildren(
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[];

  /**
   * 子エレメントを更新
   *
   * レンダリングが必要になった際に、都度 IM-Hichee 側から呼び出されるメソッドです。\
   * 引数で渡された子エレメントの配列に変更がある場合、変更後の配列を返却してください。
   *
   * 変更がない場合は、引数の `children` をそのまま返却してください。\
   * 子エレメントは配列の格納順に配置されます。
   */
  updateChildren(
    children: IUIElement[],
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[];

  /**
   * クローンされた子エレメント
   *
   * 仮想エレメントがクローンによって作成された際に IM-Hichee 側から1度だけ呼び出されるメソッドです。\
   * クローンされた子エレメントの配列が引数で渡されます。子エレメントの `parent`, `next` はこの時点では不定です。
   *
   * 子エレメントをインスタンス内で記憶させている場合は、このメソッド内で新しいインスタンスを再記憶させてください。
   *
   * 変更がない場合は、引数の `children` をそのまま返却してください。\
   * 子エレメントは配列の格納順に配置されます。
   */
  clonedChildren(
    children: IUIElement[],
    self: IUIElement,
    container: IUIContainer,
    properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>
  ): IUIElement[];
}

type UIElementDisplayType =
  /** ブロック要素 */
  | 'BLOCK'
  /** インラインブロック要素 */
  | 'INLINE'
  /** インラインフレックス要素 */
  | 'INLINE_FLEX'
  /** インライン入力要素 */
  | 'INPUT'
  /** セパレータ要素 */
  | 'SEPARATOR'
  /** テーブル要素 */
  | 'TABLE'
  /** テーブルセル要素 */
  | 'TABLE_CELL'
  /** フレックス要素 */
  | 'FLEX_BOX'
  /** フレックスアイテム要素 */
  | 'FLEX_ITEM'
  /** 無デザイン要素 */
  | 'FIXED'
  /** 選択可能な無デザイン要素 */
  | 'SELECTABLE_FIXED'
  /** 機能的な無デザイン要素 */
  | 'FUNCTIONAL'
  /** 機能を限定したインライン要素 */
  | 'CANVAS'
  /** display を指定しない要素 */
  | 'PURE';

type UIElementCoreClass = new () => IUIElementCore;
type UIElementWrapperClass =
  /** インデックスを持つエレメント */
  | 'IndexableUIElement'
  /** 繰り返すエレメント */
  | 'RepeatableUIElement'
  /** シンプルな通常のエレメント */
  | 'SimpleUIElement';

interface IIndexableUIElementCore extends IUIElementCore {}

interface IRepeatableUIElementCore extends IUIElementCore {
  /**
   * 繰り返し回数
   *
   * 子エレメントを繰り返す回数を返却します。\
   * ここで返却した回数分、自動的に子エレメントが複製されます。
   */
  repeatCount(properties: IUIElementPropertyAccessor<IUniquePropertyDefinition & ICommonPropertyDefinition>): number;
}

interface IUIElementProperties extends IUIComponentProperties {
  /**
   * エレメントID
   *
   * 画面上でエレメントを特定するための一意なIDです。\
   * 代入するとエレメントのID属性を変更できますが、重複チェックは行いませんので、重複しないように注意してください。
   */
  elementId: string;
}

interface IUIElementPropertyAccessor<DEFINITION> {
  /**
   * プロパティの値を取得
   *
   * プロパティに設定されている値を `IArgument` クラスで取得します。\
   * プロパティに変数値が設定されている場合は、最終的に解決された `IArgument` を返却します。
   */
  getProperty<K extends keyof DEFINITION>(key: K): IArgument;

  /**
   * プロパティの値を設定
   *
   * `value` に `IArgument` 以外の型を指定し、かつプロパティが変数値の場合は、変数値に代入します。\
   * 上記以外の場合は、プロパティ自体の値を変更します。
   *
   * `value` に指定した値が未確定（入力途中等）の場合は、`fixed` を `false` に指定してください。\
   * 確定済みの場合は、`fixed` を `true` に指定してください。
   */
  setProperty<K extends keyof DEFINITION>(key: K, value: CSJSPrimitiveAllType, fixed: boolean): void;

  /**
   * プロパティ値に対するバリデーションエラーメッセージを設定
   *
   * デザイナのプロパティエディタに、バリデーションエラーメッセージを表示します。\
   * デザイナの標準機能でメッセージが設定されている場合、`force` に `false` を指定した場合は、そちらが優先されます。\
   * `force` に `true` を指定した場合は、このメソッドで指定したメッセージが強制的に表示されます。\
   * 実行画面では、ブラウザのコンソールにメッセージを出力します。
   */
  setValidationError<K extends keyof DEFINITION>(key: K, message: string, force: boolean): void;

  /**
   * プロパティ値に対するバリデーションエラーメッセージをクリア
   *
   * デザイナのプロパティエディタから、バリデーションエラーメッセージを削除します。\
   * 実行画面では、何も行いません。
   */
  clearValidationError<K extends keyof DEFINITION>(key: K): void;
}

interface IUniquePropertyDefinition {
  [key: string]: UniquePropertyDefinitionType;
}

/**
 * コンポーネントのプロパティの説明は、`SIT.Z.IWP.HICHEE.COMPONENT.PROPERTY.DESCRIPTION.${コンポーネント名}.${プロパティのキー}` のメッセージプロパティキーで指定します。
 * メッセージプロパティキーが存在しない場合は、プロパティのヘルプアイコンを表示しません。
 */
type UniquePropertyDefinitionType = {
  /** 表示名 */
  readonly displayName: string;

  /** 省略可能な定義 */
  readonly definition: {
    /** バリデーションルール - 必須（未指定の場合： `false`） */
    readonly required?: boolean;

    /** バリデーションルール - データ種別（未指定の場合： `ANY`） */
    readonly dataType?: UIComponentPropertyDataType;

    /** バリデーションルール - 数値の最小値 */
    readonly min?: number;

    /** バリデーションルール - 数値の最大値 */
    readonly max?: number;

    /** 読み取り専用（未指定の場合： `false`） */
    readonly readonly?: boolean;

    /** 値変更時、再レンダリング（未指定の場合： `false`） */
    readonly rerender?: boolean;

    /** エレメント繰り返し使用（未指定の場合： `false`） */
    readonly repeatable?: boolean;

    /** プレビュー時および実行画面でのバリデーション実行（未指定の場合： `false`） */
    readonly validateValue?: boolean;

    /** 固定値許可（未指定の場合： `true`） */
    readonly acceptStatic?: boolean;

    /** 変数値許可（未指定の場合： `true`） */
    readonly acceptDynamic?: boolean;

    /** 複数行文字列の入力許可（未指定の場合： `false`） */
    readonly acceptMultipleLines?: boolean;

    /** 値の選択候補 */
    readonly candidate?: {
      /** 表示名 */
      readonly displayName: string;

      /** 値 */
      readonly value: UIComponentPropertyValueType;
    }[];
  };

  /** プロパティタイプ名 */
  readonly type: UIComponentPropertyType;

  /** プロパティ値 */
  readonly value?: {};

  /** カテゴリID */
  readonly categoryId?: string;
};

interface ICommonPropertyDefinition {
  /** onMouseDown イベント */
  readonly onMouseDown?: EventPropertyDefinitionValueType;

  /** onMouseMove イベント */
  readonly onMouseMove?: EventPropertyDefinitionValueType;

  /** onMouseEnter イベント */
  readonly onMouseEnter?: EventPropertyDefinitionValueType;

  /** onMouseLeave イベント */
  readonly onMouseLeave?: EventPropertyDefinitionValueType;

  /** onMouseOut イベント */
  readonly onMouseOut?: EventPropertyDefinitionValueType;

  /** onMouseOver イベント */
  readonly onMouseOver?: EventPropertyDefinitionValueType;

  /** onMouseUp イベント */
  readonly onMouseUp?: EventPropertyDefinitionValueType;

  /** onKeyDown イベント */
  readonly onKeyDown?: EventPropertyDefinitionValueType;

  /** onKeyPress イベント */
  readonly onKeyPress?: EventPropertyDefinitionValueType;

  /** onKeyUp イベント */
  readonly onKeyUp?: EventPropertyDefinitionValueType;

  /** onFocus イベント */
  readonly onFocus?: EventPropertyDefinitionValueType;

  /** onBlur イベント */
  readonly onBlur?: EventPropertyDefinitionValueType;

  /** onChange イベント */
  readonly onChange?: EventPropertyDefinitionValueType;

  /** onInput イベント */
  readonly onInput?: EventPropertyDefinitionValueType;
}

type EventPropertyDefinitionValueType =
  /** このプロパティを使用する */
  'use';

interface IUIElementConstructorArguments {
  /** コンテナ */
  readonly container: IUIContainer;

  /** 親のエレメント */
  readonly parent: ParentUIElement;

  /** 次のエレメント */
  readonly next: NextUIElement;

  /** UUID */
  readonly uuid: string | null;

  /** エレメントID */
  readonly elementId: string | null;

  /** 元のエレメント */
  readonly source: IUIElement | null;

  /** 子エレメントの配列 */
  readonly children: IUIElement[] | null;

  /** 子エレメント自動クローン */
  readonly cloneChildren: boolean;
}

interface IUIElementRepository {
  /** エレメントクラスの一覧 */
  readonly classes: UIElementCoreClass[];

  /** エレメントカテゴリの一覧 */
  readonly categories: UIElementCategoryInfo[];

  /** エレメントクラスを登録 */
  registerClass(clazz: UIElementCoreClass, options?: UIElementRegisterOptions): void;

  /** エレメントカテゴリを登録 */
  registerCategory(categoryId: string, options?: UIElementCategoryRegisterOptions): void;

  /** エレメントクラスを取得 */
  getClass(elementTypeName: string): UIElementCoreClass;

  /** エレメントクラスのオプションを取得 */
  getClassOptions(clazz: UIElementCoreClass): UIElementRegisterOptions;

  /** エレメントクラスのオプションを設定 */
  setClassOptions(clazz: UIElementCoreClass, options: UIElementRegisterOptions): void;

  /** エレメントカテゴリのオプションを取得 */
  getCategoryOptions(categoryId: string): UIElementCategoryInfo;

  /** エレメントカテゴリのオプションを設定 */
  setCategoryOptions(categoryId: string, options: UIElementCategoryRegisterOptions): void;
}

/**
 * コンポーネントのカテゴリ名は、`CAP.Z.IWP.HICHEE.COMPONENT.CATEGORY.NAME.{カテゴリID}` のメッセージプロパティキーで指定します。
 * このキーが存在しない場合はカテゴリIDを使用します。
 */
type UIElementCategoryInfo = {
  /** カテゴリID */
  readonly categoryId: string;

  /** オプション */
  readonly options?: UIElementCategoryRegisterOptions;
};

type UIElementCategoryRegisterOptions = {
  /** ソート番号（昇順） */
  readonly sortNumber?: number;
};

/**
 * コンポーネント名は、`CAP.Z.IWP.HICHEE.COMPONENT.NAME.{クラス名}` のメッセージプロパティキーで指定します。
 * このキーが存在しない場合はクラス名を使用します。
 */
type UIElementRegisterOptions = {
  /** カテゴリID */
  readonly categoryId?: string;

  /** ソート番号 */
  readonly sortNumber?: number;

  /**
   * パレットでの非表示フラグ
   *
   * このフラグは、デザイナのパレットによって書き換えられます。
   */
  readonly hiddenInPalette?: boolean;
};

interface IUIElementIDResolver {
  /** エレメントIDを採番 */
  createElementId(): string;

  /** エレメントIDの枝番を取得 */
  createElementIdWithSuffixCode(element: IUIElement): string;

  /**
   * エレメントIDを登録
   *
   * 重複している場合は、新しいエレメントIDを返却します。
   */
  registerElementId(element: IUIElement): string;

  /** エレメントIDを解放 */
  releaseElementId(element: IUIElement): void;

  /** クリア */
  clear(): void;
}

interface IUIElementController {
  /**
   * インスタンスの配列
   *
   * このエレメントコントローラが管理している全てのエレメントを取得します。
   */
  readonly instances: IUIElement[];

  /**
   * エレメントのインスタンスを作成
   *
   * エレメントが持つ `elementTypeName` を指定して、エレメントのインスタンスを作成します。
   */
  createInstanceByTypeName(elementTypeName: string, parent: ParentUIElement, next?: NextUIElement): IUIElement;

  /**
   * エレメントのインスタンスを作成
   *
   * エレメントクラスを指定して、エレメントのインスタンスを作成します。
   */
  createInstanceByClass<T extends IUIElement>(
    clazz: UIElementCoreClass,
    parent: ParentUIElement,
    next?: NextUIElement
  ): T;

  /**
   * エレメントのインスタンスを複製
   *
   * エレメントのインスタンスを複製します。\
   * `parent` で複製する先を指定できます。 指定しない場合、複製元と同じ親に複製します。\
   * `cloneChildren` に `true` を指定した場合、子エレメントも再帰で複製します。
   */
  cloneInstance(source: IUIElement, parent?: ParentUIElement, cloneChildren?: boolean): IUIElement;

  /**
   * エレメントのインスタンスをデシリアライズ
   *
   * エレメントの UUID を指定して、エレメントのインスタンスを作成します。\
   * このメソッドはシリアライズ化されたエレメントのオブジェクトを元に戻す場合に使用します。
   */
  deserializeInstance(elementTypeName: string, parent: ParentUIElement, uuid: string): IUIElement;

  /**
   * エレメントのインスタンス属性と子エレメントを上書き
   *
   * `source` から `target` へ、エレメントの情報をコピーします。\
   * 引き継ぐ情報には、全てのプロパティ、子エレメントが含まれます。
   */
  overwriteInstance(source: IUIElement, target: IUIElement): IUIElement[];

  /**
   * UUID からエレメントのインスタンスを取得
   *
   * このエレメントコントローラが管理している全てのエレメントから、UUID が一致するエレメントを取得します。\
   * 一致するエレメントが存在しない場合、`null` を返却します。
   */
  getInstanceByUUID(uuid: string): IUIElement | null;

  /**
   * 親エレメントに属する子エレメントの全インスタンスを取得
   *
   * 指定したエレメントの配下にある子エレメントを、1階層分取得します。\
   * 全ての階層にある子エレメントを再帰的に取得する場合、`getChildrenDeep` メソッドを使用してください。
   */
  getChildren(parent: ParentUIElement): IUIElement[];

  /**
   * 親エレメントに属する子孫エレメントの全インスタンスを取得
   *
   * 指定したエレメントの配下にある子エレメントを、再帰的に取得します。\
   * 全ての階層にある子エレメントを全て取得するため、実行に時間がかかります。
   */
  getChildrenDeep(parent: ParentUIElement): IUIElement[];

  /**
   * 親エレメントに属する子エレメントの全インスタンスを配置順に取得
   *
   * 指定したエレメントの配下にある子エレメントを、コンテナ上で描画される順序で取得します。\
   * `IUIElement#children` プロパティでは順序が保証されていませんが、このメソッドで返却される子エレメントの順序は保証されます。
   */
  getSortedChildren(parent: ParentUIElement): IUIElement[];

  /**
   * 親エレメントの全インスタンスを取得
   *
   * 指定したエレメントより親のコンポーネントを全て取得します。\
   * 取得されるコンポーネントには、エレメントとコンテナが含まれます。
   */
  getParents(child: IUIElement): ParentUIElement[];
}

/* ********************************************************************** *
 * アクション（共通）
 * ********************************************************************** */

interface IUIContainerActionDisplayCore {
  /** アクションの本文を示すメッセージキー */
  readonly messageKey: string;

  /** パラメータ定義 */
  readonly parameterDefinition: IUIContainerActionParameterDefinition;
}

interface IUIContainerActionParameterDefinition {
  [key: string]: UIContainerActionParameterDefinitionType;
}

type UIContainerActionParameterDefinitionType = {
  /** パラメータ定義タイプ */
  readonly type: UIContainerActionParameterCandidateType;

  /** パラメータ名を示すメッセージキー */
  readonly messageKey: string;

  /**
   * パラメータの選択候補のラベル
   *
   * messageKey が指定されている場合は messageKey を使用します。\
   * messageKey が指定されていない、または該当の messageKey が存在しない場合は、displayName を使用します。\
   * どちらも存在しない場合は、value の値を使用します。
   */
  readonly candidate?: {
    readonly messageKey?: string;
    readonly displayName?: string;
    readonly value: string;
  }[];
};

type UIContainerActionParameterCandidateType =
  /** 定数・入力 */
  | 'variable-input'
  /** 変数 */
  | 'variable-output'
  /** 変数・定数・入力 */
  | 'variable-all'
  /** 直接入力 */
  | 'literal'
  /** 変数・直接入力 */
  | 'variable-literal'
  /** ラベル定義 */
  | 'label-target'
  /** アクション内ラベル */
  | 'label'
  /** アクション（自分を除く） */
  | 'action'
  /** コンテナページ */
  | 'page'
  /** テーブル */
  | 'table-select'
  /** チェックボックス */
  | 'checkbox'
  /** JavaScript エディタ */
  | 'javascript'
  /** 隠しパラメータ */
  | 'hidden'
  /** ロジックフロールーティング */
  | 'im-logic-flow-route'
  /** ViewCreatorルーティング */
  | 'viewcreator-route';

interface IUIContainerActionObjectCollector {}

interface ITableSelectActionParameter extends IUIContainerActionParameter {
  /** テーブルカラム形式で取得 */
  toColumns(container: IUIContainer, component?: IUIComponent): UIContainerActionDialogColumn[];
}

type UIContainerActionDialogColumn = {
  readonly key: string;
  value: string; // 一時変数扱いのため、書き換え可能
  staticValue?: string; // 固定値
  dynamicValue?: string; // 変数値
};

/**
 * アクションアイテムのカテゴリ名は、`CAP.Z.IWP.HICHEE.ACTION.ITEM.CATEGORY.NAME.{任意のキー}` のメッセージプロパティキーで指定します。
 * アクションアイテムの実行条件のカテゴリ名は、`CAP.Z.IWP.HICHEE.ACTION.CONDITION.CATEGORY.NAME.{任意のキー}` のメッセージプロパティキーで指定します。
 */
type UIContainerActionCategoryInfo = {
  /** カテゴリID */
  readonly categoryId: string;

  /** オプション */
  readonly options?: UIContainerActionCategoryRegisterOptions;
};

type UIContainerActionCategoryRegisterOptions = {
  /** ソート番号（昇順） */
  readonly sortNumber?: number;
};

/* ********************************************************************** *
 * アクションチャート
 * ********************************************************************** */

interface IActionChartStorage {
  /** アクションチャート一覧 */
  readonly charts: IUIContainerActionChart[];

  /** アクションチャートを取得 */
  getChart(uuid: string): IUIContainerActionChart | null;

  /** アクションコンテキストを作成 */
  createContext(chart: IUIContainerActionChart, event: UIContainerActionContextEventType): IUIContainerActionContext;
}

interface IUIContainerActionChart {
  /** UUID */
  readonly uuid: string;

  /** アクションチャート名 */
  readonly name: string;

  /** アクションが呼ばれたときのアクションアイテム（正常系） */
  readonly items: IUIContainerActionItem[];

  /** エラーが発生したときのアクションアイテム（異常系） */
  readonly errorItems: IUIContainerActionItem[];

  /**
   * アクションアイテムの処理を順次実行
   *
   * 全ての正常系アクションアイテムとアクション実行条件の処理を、順次実行します。\
   * アクションアイテムに紐づけられたアクション実行条件判定処理を行い、処理実行判定となったアクションアイテムを実行します。
   *
   * アクションアイテム、実行条件ともに、1つずつ処理を実行することを保証します。\
   * したがって、アクションアイテム、実行条件が同期型・非同期型にかかわらず、配列の格納順に実行します。\
   * 非同期型で実装されている場合は、処理完了まで待機します。
   *
   * このメソッドは非同期での実行を想定しています。\
   * メソッドを使用する際には `await` または、待機処理を適切に実装してください。
   */
  run(context: IUIContainerActionContext, component: IUIComponent): Promise<void>;

  /** インスタンスを複製 */
  clone(container: IUIContainer): IUIContainerActionChart;
}

interface IUIContainerActionContext {
  /** アクションチャート */
  readonly chart: IUIContainerActionChart;

  /** ブラウザイベント */
  readonly event: UIContainerActionContextEventType;

  /**
   * アクションが呼ばれたときに実行するアクションアイテムのインデックス
   *
   * IUIContainerActionChart#items で何番目のアクションアイテムを実行しているかを示します。\
   * この値は、アクションの実行中に IM-BloomMaker 本体によって書き換えられます。
   */
  calledActionIndex: number;

  /**
   * エラーが発生したときに実行するアクションアイテムのインデックス
   *
   * IUIContainerActionChart#errorItems で何番目のアクションアイテムを実行しているかを示します。\
   * この値は、アクションの実行中に IM-BloomMaker 本体によって書き換えられます。
   */
  errorOccurredActionIndex: number;

  /**
   *  「アクションが呼ばれたとき」に定義したアクションでエラーが発生したかどうかを示すフラグです。
   */
  hasErrorOnCalledAction: boolean;

  /**
   * 「エラーが発生したとき」に定義したアクションでエラーが発生したかどうかを示すフラグです。
   */
  hasErrorOnErrorOccurredAction: boolean;
}

type UIContainerActionContextEventType = MouseEvent | KeyboardEvent | null;

/* ********************************************************************** *
 * アクションアイテム
 * ********************************************************************** */

interface IUIContainerActionItem {
  /** UUID */
  readonly uuid: string;

  /** アクション有効 */
  enabled: boolean;

  /** コメント */
  comment: string;

  /** アクション発生条件 */
  conditions: IUIContainerActionCondition[];

  /** インスタンスを複製 */
  clone(): IUIContainerActionItem;
}

interface IUIContainerActionCondition {
  /** UUID */
  readonly uuid: string;

  /** インスタンスを複製 */
  clone(): IUIContainerActionCondition;
}

interface IUIContainerActionItemCore extends IUIContainerActionDisplayCore {
  /**
   * アクションタイプ名
   *
   * アクションアイテムのクラス名と同一の文字列を取得します。
   */
  readonly actionTypeName: string;

  /**
   * アクションアイテムの処理を実行
   *
   * アクションアイテムの処理を実行します。\
   * 処理が同期型か非同期型かにかかわらず、`Promise` を返却することを保証します。
   *
   * このメソッドは非同期での実行を想定しています。\
   * メソッドを使用・実装する際には `await` または、待機処理を適切に実装してください。
   */
  run(
    context: IUIContainerActionContext,
    container: IUIContainer,
    component: IUIComponent,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>
  ): Promise<void>;

  /**
   * 動的に変更されるパラメータの値を生成
   *
   * アクションに定義されているパラメータのうち、画面上に表示されずアクション側で補完するための値を作成します。\
   * パラメータ値を補完する必要がない場合は、このメソッドを実装する必要はありません。
   */
  createDynamicParameterValue?(
    container: IUIContainer,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>,
    key: string
  ): string;
}

interface IUIContainerActionItemRepository {
  /** アクションアイテムクラス一覧 */
  readonly classes: UIContainerActionItemClass[];

  /** アクションアイテムのカテゴリ一覧 */
  readonly categories: UIContainerActionCategoryInfo[];

  /** アクションアイテムクラスを登録 */
  registerClass(clazz: UIContainerActionItemClass, options?: UIContainerActionRegisterOptions): void;

  /** アクションアイテムのカテゴリを登録 */
  registerCategory(categoryId: string, options?: UIContainerActionCategoryRegisterOptions): void;

  /** アクションアイテムクラスを取得 */
  getClass(actionTypeName: string): UIContainerActionItemClass;

  /** アクションアイテムクラスのオプションを取得 */
  getClassOptions(clazz: UIContainerActionItemClass): UIContainerActionRegisterOptions;

  /** アクションアイテムクラスのオプションを設定 */
  setClassOptions(clazz: UIContainerActionItemClass, options: UIContainerActionRegisterOptions): void;

  /** アクションアイテムのカテゴリのオプションを取得 */
  getCategoryOptions(categoryId: string): UIContainerActionCategoryInfo;

  /** アクションアイテムのカテゴリのオプションを設定 */
  setCategoryOptions(categoryId: string, options: UIContainerActionCategoryRegisterOptions): void;
}

type UIContainerActionItemClass = new () => IUIContainerActionItemCore;

/**
 * アクションアイテム名は、`CAP.Z.IWP.HICHEE.ACTION.ITEM.LABEL.{任意のキー}` のメッセージプロパティキーで指定します。\
 * アクションアイテムの実行条件名は、`CAP.Z.IWP.HICHEE.ACTION.CONDITION.PARAMETER.CONTAINER.{任意のキー}` のメッセージプロパティキーで指定します。
 */
type UIContainerActionRegisterOptions = {
  /** カテゴリID */
  readonly categoryId?: string;

  /** ソート番号 */
  readonly sortNumber?: number;

  /**
   * パレットでの非表示フラグ
   *
   * このフラグは、デザイナのパレットによって書き換えられます。
   */
  readonly hiddenInPalette?: boolean;
};

/* ********************************************************************** *
 * アクション実行条件
 * ********************************************************************** */

interface IUIContainerActionConditionRepository {
  /** アクション条件クラスの一覧 */
  readonly classes: UIContainerActionConditionClass[];

  /** アクション条件のカテゴリ一覧 */
  readonly categories: UIContainerActionCategoryInfo[];

  /** アクション条件クラスを登録 */
  registerClass(clazz: UIContainerActionConditionClass, options?: UIContainerActionRegisterOptions): void;

  /** アクション条件のカテゴリを登録 */
  registerCategory(categoryId: string, options?: UIContainerActionCategoryRegisterOptions): void;

  /** アクション条件クラスを取得 */
  getClass(actionTypeName: string): UIContainerActionConditionClass;

  /** アクション条件クラスのオプションを取得 */
  getClassOptions(clazz: UIContainerActionConditionClass): UIContainerActionRegisterOptions;

  /** アクション条件クラスのオプションを設定 */
  setClassOptions(clazz: UIContainerActionConditionClass, options: UIContainerActionRegisterOptions): void;

  /** アクション条件のカテゴリのオプションを取得 */
  getCategoryOptions(categoryId: string): UIContainerActionCategoryInfo;

  /** アクション条件のカテゴリのオプションを設定 */
  setCategoryOptions(categoryId: string, options: UIContainerActionCategoryRegisterOptions): void;
}

interface IUIContainerActionConditionCore extends IUIContainerActionDisplayCore {
  /**
   * 条件タイプ名
   *
   * アクション実行条件のクラス名と同一の文字列を取得します。
   */
  readonly conditionTypeName: string;

  /**
   * 条件判定を実行
   *
   * アクション実行条件の処理を実行して、判定結果を返却します。\
   * 与えられた条件を満たす場合は `true`、満たさない場合は `false` を `Promise` で返却します。
   *
   * このメソッドは非同期での実行を想定しています。\
   * メソッドを使用・実装する際には `await` または、待機処理を適切に実装してください。
   */
  matched(
    context: IUIContainerActionContext,
    container: IUIContainer,
    component: IUIComponent,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>
  ): Promise<boolean>;

  /**
   * 動的に変更されるパラメータの値を生成
   *
   * 条件に定義されているパラメータのうち、画面上に表示されず条件側で補完するための値を作成します。\
   * パラメータ値を補完する必要がない場合は、このメソッドを実装する必要はありません。
   */
  createDynamicParameterValue?(
    container: IUIContainer,
    parameters: IUIContainerActionParameterAccessor<IUIContainerActionParameterDefinition>,
    key: string
  ): string;
}

type UIContainerActionConditionClass = new () => IUIContainerActionConditionCore;

/* ********************************************************************** *
 * アクションパラメータ
 * ********************************************************************** */

interface IUIContainerActionParameterAccessor<DEFINITION> {
  /** プロパティを取得 */
  getParameter<K extends keyof DEFINITION>(key: K): IUIContainerActionParameter;
}

interface IUIContainerActionParameter {
  /** プロパティ設定値で取得 */
  toRaw(): string;

  /** プリミティブ値で取得 */
  toPrimitive(container: IUIContainer): CSJSPrimitiveAllType;

  /** Argument で取得 */
  toArgument(container: IUIContainer, component?: IUIComponent): IArgument;
}

/* ********************************************************************** *
 * Parameter
 * ********************************************************************** */

interface IParameterStorage {
  /**
   * パスが存在するかどうかを取得
   *
   * 引数 component を基準として指定したパスが存在するかどうかを取得します。
   */
  existsPath(component: IUIComponent, path: string): boolean;

  /**
   * パス指定でパラメータを受け取り、その配下のパラメータから name を配列で取得
   *
   * 引数 component を基準として指定したパスからパラメータを受け取り、その配下のパラメータから name を配列で取得します。
   */
  getParameterChildrenName(component: IUIComponent, path: string): string[];

  /**
   * Argument を読み取り
   *
   * 引数 component を基準として指定したパスに紐づく変数の Argument を取得します。\
   * パスが存在しない場合は null を示す Argument を返却します。
   */
  read(component: IUIComponent, path: string): IArgument;

  /**
   * 親子関係を含む全ての Argument を読み取り
   *
   * 引数 component を基準として指定したパスに紐づく変数の Argument を全て取得します。\
   * 返却された配列には、パスの親からたどり、解決された順に Argument が格納されます。
   */
  readAllChain(component: IUIComponent, path: string): IArgument[];

  /**
   * Argument を書き込み
   *
   * 引数 component を基準として指定したパスに紐づく変数に Argument を上書きします。\
   * 元々パスに紐づく変数に格納されていた Argument を返却します。\
   * パスが存在しない場合は null を示す Argument を返却します。
   */
  writeAsArgument(component: IUIComponent, path: string, argument: IArgument): IArgument;

  /**
   * パラメータを上書きして Argument を書き込み
   *
   * 引数 component を基準として指定したパスに紐づく変数に Argument を上書きします。\
   * 元々パスに紐づく変数に格納されていた Argument を返却します。\
   * パスに紐づく変数のパラメータを、指定した Argument の型で上書きします。
   */
  overwriteAsArgument(component: IUIComponent, path: string, argument: IArgument): IArgument;

  /**
   * プリミティブ値を書き込み（配列考慮なし）
   *
   * 引数 component を基準として指定したパスに紐づく変数にプリミティブ値を上書きします。\
   * プリミティブ値は変数のパラメータに指定されている型情報に応じて、自動的に対応する Argument に変換されます。
   */
  writeAsPrimitive(component: IUIComponent, path: string, value: CSJSPrimitiveAllType): IArgument;

  /**
   * プリミティブ値を書き込み（配列考慮あり）
   *
   * writeAsPrimitive メソッドと同等の型変換を提供しますが、このメソッドはさらに配列を考慮します。\
   * 変数のパラメータが配列の場合、自動的に要素数1の ArrayArgument に変換されます。\
   * 変数のパラメータが配列でなく、かつ value に配列が指定された場合、最初の要素に対して Argument に変換されます。
   */
  writeAsPrimitiveStrictArrayTyped(component: IUIComponent, path: string, value: CSJSPrimitiveAllType): IArgument;

  /**
   * Argument をプリミティブに変換
   *
   * 引数の Argument を CSJS で扱えるプリミティブ値に変換します。\
   * DateArgument は Date クラスに変換します。
   */
  toPrimitive(argument: IArgument): CSJSPrimitiveAllType;

  /**
   * Argument を JSON オブジェクトに変換
   *
   * 引数の Argument を CSJS で扱える JSON オブジェクトに変換します。\
   * DateArgument は文字列に変換します。
   */
  toJSONObject(argument: IArgument): CSJSPrimitiveAllType;

  /**
   * Argument をクエリに変換
   *
   * 引数の Argument を IM-BloomMaker のサーバ側で扱えるクエリパラメータ形式に変換します。
   */
  toQueryParameters(baseName: string, argument: IArgument): QueryParametersType;

  /**
   * ObjectArgument をクエリで使用するキー・文字列値に変換
   *
   * 引数の ObjectArgument を jQuery 互換のクエリパラメータ形式を作成するためのキー・文字列値に変換します。
   */
  toQueryKeyValues(argument: IObjectArgument): QueryKeyValuesType;
}

type QueryParametersType = {[key: string]: string};
type QueryKeyValuesType = {[key: string]: string; value: string}[];

/* ********************************************************************** *
 * Argument
 * ********************************************************************** */

interface IArgument {
  /** Argument クラス */
  readonly clazz: ArgumentClass;

  /** Argument ごとに割り当てられた UUID */
  readonly uuid: string;

  /**
   * Argument タイプ名
   *
   * 以下の形式かどうかを確認する場合は、比較関数を使用してください。
   * * 配列かどうか -> `isArray()`
   * * マップ型かどうか -> `isObject()`
   * * 日付・時刻型かどうか -> `isDate()`
   * * null かどうか -> `isNull()`
   * * 有効な値かどうか -> `isValid()`
   */
  readonly argumentTypeName: ArgumentTypeName;

  /** Argument タイプ表示名 */
  readonly displayType: string;

  /** Argument タイプ表示値 */
  readonly displayValue: string;

  /** イベント優先度 */
  readonly eventPriority: number;

  /** 文字列で取得 */
  toString(): string;

  /** 数値で取得 */
  toNumber(): number;

  /** 真偽値で取得 */
  toBoolean(): boolean;

  /** 配列かどうか */
  isArray(): this is IArrayArgument;

  /** マップ型かどうか */
  isObject(): this is IObjectArgument;

  /** 日付・時刻型かどうか */
  isDate(): this is IDateArgument;

  /** null かどうか */
  isNull(): boolean;

  /**
   * 有効な値かどうか
   *
   * 有効な値とは、以下に示す値に該当しないものです。
   * * null
   * * 空の文字列
   * * NaN
   * * Invalid Date
   */
  isValid(): boolean;

  /** 一致するかどうか */
  equals(target: IArgument): boolean;

  /** Argument の複製 */
  clone(): IArgument;
}

interface IPrimitiveArgument extends IArgument {}

interface IChainedArgument extends IPrimitiveArgument {
  /** 解決された Argument */
  readonly resolved: IArgument;

  /** リフレッシュ */
  refresh(): void;
}

interface IDateArgument extends IPrimitiveArgument {
  /** 日時情報 */
  raw: IDateTime;
}

interface II18nArgument extends IPrimitiveArgument {
  /** ロケールIDの配列を取得 */
  getLocaleIds(): string[];

  /** ロケールIDの存在チェック */
  containsLocaleId(localeId: string): boolean;

  /** Argument を取得 */
  get(localeId: string): string | null;

  /** Argument を設定 */
  put(localeId: string, message: string): string | null;

  /** Argument を削除 */
  remove(localeId: string): string;
}

interface INumberArgument extends IPrimitiveArgument {}

interface IRelationArgument extends IArgument {
  /** 関連する Argument を更新 */
  updateRelatedArgument(oldArgument: IArgument, newArgument: IArgument): void;
}

interface IArrayArgument extends IRelationArgument {
  /** 配列の件数 */
  readonly length: number;

  /** 空かどうかを取得 */
  isEmpty(): boolean;

  /** 全てクリア */
  clear(): void;

  /** Argument を取得 */
  get(index: number): IArgument;

  /** Argument を設定 */
  set(index: number, value: IArgument): IArgument;

  /** Argument を列挙 */
  forEach(callback: ArrayArgumentForEachCallback): void;

  /** Argument を差し替えた配列を取得 */
  map<T>(callback: ArrayArgumentMapCallback<T>): T[];

  /** Argument を列挙して true 返却があるかどうか */
  some(callback: ArrayArgumentSomeCallback): boolean;

  /** Argument を末尾に追加 */
  push(...values: IArgument[]): void;

  /** Argument の内容を追加 */
  pushAll(value: IArrayArgument): void;

  /** Argument を末尾から取り出す */
  pop(): IArgument;

  /** Argument を先頭に追加 */
  unshift(...value: IArgument[]): void;

  /** Argument を先頭から取り出す */
  shift(): IArgument;

  /** Argument を挿入 */
  insert(index: number, value: IArgument): void;

  /** Argument の内容を挿入 */
  insertAll(index: number, value: IArrayArgument): void;

  /** Argument を削除 */
  remove(index: number, size: number): IArgument;

  /** Argument の内容を総入れ替え */
  replaceAll(value: IArrayArgument): void;
}

type ArrayArgumentForEachCallback = (value: IArgument, index: number, array: IArrayArgument) => void;
type ArrayArgumentMapCallback<T> = (value: IArgument, index: number, array: IArrayArgument) => T;
type ArrayArgumentSomeCallback = (value: IArgument, index: number, array: IArrayArgument) => boolean;

interface IObjectArgument extends IRelationArgument {
  /** 配列の件数 */
  readonly length: number;

  /** 空かどうかを取得 */
  isEmpty(): boolean;

  /** 全てクリア */
  clear(): void;

  /** キーの配列を取得 */
  getKeys(): string[];

  /** キーの存在チェック */
  containsKey(key: string): boolean;

  /** Argument を取得 */
  get(key: string): IArgument;

  /** Argument を設定 */
  put(key: string, value: IArgument): IArgument;

  /** Argument の内容を設定 */
  putAll(value: IObjectArgument): void;

  /** Argument を列挙 */
  forEach(callback: ObjectArgumentForEachCallback): void;

  /** Argument を差し替えた配列を取得 */
  map<T>(callback: ObjectArgumentMapCallback<T>): T[];

  /** Argument を列挙して true 返却があるかどうか */
  some(callback: ObjectArgumentSomeCallback): boolean;

  /** Argument を削除 */
  remove(key: string): IArgument;

  /** キーを入れ替え */
  replaceKey(sourceKey: string, targetKey: string): void;

  /** Argument の内容を総入れ替え */
  replaceAll(value: IObjectArgument): void;
}

type ObjectArgumentForEachCallback = (value: IArgument, key: string, object: IObjectArgument) => void;
type ObjectArgumentMapCallback<T> = (value: IArgument, key: string, object: IObjectArgument) => T;
type ObjectArgumentSomeCallback = (value: IArgument, key: string, object: IObjectArgument) => boolean;

interface IPropertyValueArgument extends IChainedArgument {
  /**
   * この Argument がラップする生の値
   *
   * ユーザが指定したプロパティ値の生データを取得・設定するためにのみ使用します。\
   * 実際に実行画面で使われるプロパティ値を取得する場合は raw を使用せず、
   * 直接 `toString`, `toNumber`, `toBoolean` メソッドを使用してください。
   */
  raw: IArgument;

  /** Argument を監視 */
  observe(target: IUIComponent, property: IUIComponentProperty): void;

  /** Argument を解放 */
  release(): void;
}

interface IParameterPathArgument extends IChainedArgument {
  /**
   * この Argument がラップする生の値
   *
   * ユーザが指定したプロパティ値の生データを取得・設定するためにのみ使用します。\
   * 実際に実行画面で使われるプロパティ値を取得する場合は raw を使用せず、
   * 直接 `toString`, `toNumber`, `toBoolean` メソッドを使用してください。
   */
  raw: string;

  /** ターゲットとなるコンポーネント */
  readonly target: IUIComponent;

  /** Argument を監視 */
  observe(): void;

  /** Argument を解放 */
  release(): void;
}

interface IDateTime {
  /** 日時オブジェクト */
  readonly date: Date;

  /** 時差（ms） */
  readonly offset: number;

  /** 文字列化 */
  toString(): string;

  /** 数値化（エポックミリ秒） */
  toNumber(): number;

  /** ISO8601 形式で文字列化 */
  toISOString(offset?: number): string;

  /** ローカル時差の Date に変換 */
  toLocalDate(): Date;

  /** 有効な日時かどうか */
  isValid(): boolean;

  /** 成分分割 */
  getComponents(): DateTimeComponents;

  /** インスタンスを複製 */
  clone(): IDateTime;

  /** 一致比較 */
  equals(target: IDateTime): boolean;

  /** 差分比較 */
  compare(target: IDateTime): number;
}

type DateTimeComponents = {
  /** 日付 */
  readonly date: {
    readonly iso: string;
    readonly year: number;
    readonly month: number; // 1-12
    readonly day: number;
  };

  /** 時刻 */
  readonly time: {
    readonly iso: string;
    readonly hours: number;
    readonly minutes: number;
    readonly seconds: number;
    readonly milliseconds: number;
  };

  /** 時差 */
  readonly offset: {
    readonly iso: string;
    readonly milliseconds: number;
  };
};

interface IFraction {
  /** 分子 */
  readonly numerator: number;

  /** 分母 */
  readonly denominator: number;

  /** 正の数かどうか */
  readonly plus: boolean;

  /** ゼロかどうか */
  readonly zero: boolean;

  /** 負の数かどうか */
  readonly minus: boolean;

  /** 割り切れるかどうか */
  readonly divisible: boolean;

  /** 文字列化 */
  toString(): string;

  /** 数値化（精度低下注意） */
  toNumber(): number;

  /** インスタンスを複製 */
  clone(): IFraction;

  /** 一致比較 */
  equals(target: IFraction): boolean;

  /** 差分比較 */
  compare(target: IFraction): number;

  /** 加算 */
  add(num: number | string | IFraction, denom?: number): IFraction;

  /** 減算 */
  sub(num: number | string | IFraction, denom?: number): IFraction;

  /** 乗算 */
  mul(num: number | string | IFraction, denom?: number): IFraction;

  /** 除算 */
  div(num: number | string | IFraction, denom?: number): IFraction;

  /** 余り */
  mod(num: number | string | IFraction, denom?: number): IFraction;

  /** 絶対値 */
  abs(): IFraction;

  /** 切捨て */
  floor(): number;

  /** 四捨五入 */
  round(): number;

  /** 繰上げ */
  ceil(): number;
}

interface IKeyValueMap<K extends string, V> {
  /** キーの数 */
  readonly length: number;

  /** キーの配列 */
  readonly keys: K[];

  /** 値の配列 */
  readonly values: V[];

  /** クリア */
  clear(): void;

  /** インスタンスを複製 */
  clone(): IKeyValueMap<K, V>;

  /** キーか空かどうか */
  isEmpty(): boolean;

  /** 繰り返し */
  each(callback: (key: K, value: V) => boolean | void): number;

  /** true を返却した要素のみ取り出す */
  filter(callback: (key: K, value: V) => boolean): IKeyValueMap<K, V>;

  /** キーが含まれているかどうか */
  containsKey(key: K): boolean;

  /** 値が含まれているかどうか */
  containsValue(value: V): boolean;

  /** 値を取得 */
  get(key: K): V;

  /** 値を設定 */
  put(key: K, value: V): V;

  /** キーを削除 */
  remove(key: K): V;
}

type ArgumentClass =
  /** 配列 */
  | 'ArrayArgument'
  /** 真偽値 */
  | 'BooleanArgument'
  /** 日付 */
  | 'DateArgument'
  /** 浮動小数 */
  | 'DoubleArgument'
  /** 式表現 */
  | 'FormulaArgument'
  /** 高精度小数 */
  | 'FractionArgument'
  /** 多言語 */
  | 'I18nArgument'
  /** 整数 */
  | 'IntegerArgument'
  /** オブジェクト・ハッシュマップ */
  | 'ObjectArgument'
  /** null */
  | 'NullArgument'
  /** 変数パス */
  | 'ParameterPathArgument'
  /** プロパティ値 */
  | 'PropertyValueArgument'
  /** 文字列 */
  | 'StringArgument';

type ArgumentTypeName =
  /** 配列 */
  | 'array'
  /** 真偽値 */
  | 'boolean'
  /** 日付 */
  | 'date'
  /** 浮動小数 */
  | 'double'
  /** 式表現 */
  | 'formula'
  /** 高精度小数 */
  | 'fraction'
  /** IM-Repository 定義 */
  | 'im-repository'
  /** 多言語 */
  | 'i18n'
  /** 整数 */
  | 'integer'
  /** オブジェクト・ハッシュマップ */
  | 'object'
  /** null */
  | 'null'
  /** 変数パス */
  | 'parameter-path'
  /** プロパティ値 */
  | 'property-value'
  /** 文字列 */
  | 'string';

/* ********************************************************************** *
 * HTMLElementBuilder
 * ********************************************************************** */

interface IHTMLElementBuilder {
  /** HTML エレメントビルダのクラス */
  readonly clazz: HTMLElementBuilderClass;

  /** HTML タグ */
  readonly tag: HTMLElement;

  /** HTML エレメントビルダの設定 */
  readonly config: IHTMLElementConfig;

  /** クラス名を保持しているかどうかを取得 */
  hasClass(className: string): boolean;

  /** クラス名を追加 */
  addClass(className: string): IHTMLElementBuilder;

  /** クラス名を削除 */
  removeClass(className: string): IHTMLElementBuilder;

  /** クラス名をクリア */
  clearClass(): IHTMLElementBuilder;

  /** 引数 on によってクラス名を追加・削除 */
  switchClass(on: boolean, className: string): IHTMLElementBuilder;

  /** CSS を取得 */
  getCSS(name: string): string | null;

  /** CSS を設定 */
  setCSS(name: string, value: string): IHTMLElementBuilder;

  /** CSS を解除 */
  removeCSS(name: string): IHTMLElementBuilder;

  /** 引数 on によって CSS を設定・解除 */
  switchCSS(on: boolean, name: string, value: string): IHTMLElementBuilder;

  /** 位置を指定する CSS 設定のショートハンド */
  setPosition(position: {top?: number; right?: number; bottom?: number; left?: number}): IHTMLElementBuilder;

  /** サイズを指定する CSS 設定のショートハンド */
  setSize(size: {width?: string | number; height?: string | number}): IHTMLElementBuilder;

  /** 属性を取得 */
  getAttribute(name: string): string | null;

  /**
   * 属性を設定
   *
   * `'style'` を変更する場合は、`setCSS` メソッドを使用してください。\
   * `'class'` を変更する場合は、`addClass` メソッドを使用してください。
   */
  setAttribute(name: string, value?: string): IHTMLElementBuilder;

  /**
   * 属性を削除
   *
   * `'style'` を削除する場合は、`removeCSS` メソッドを使用してください。\
   * `'class'` を削除する場合は、`clearClass` メソッドを使用してください。
   */
  removeAttribute(name: string): IHTMLElementBuilder;

  /** 引数 on によって属性を設定・解除 */
  switchAttribute(on: boolean, name: string, value?: string): IHTMLElementBuilder;

  /**
   * タグ値を取得
   *
   * このエレメントビルダで管理するタグの value 属性値を取得します。
   */
  getValue(): string | null;

  /**
   * タグ値を設定
   *
   * このエレメントビルダで管理するタグの value 属性値を設定します。
   */
  setValue(value: string | null): IHTMLElementBuilder;

  /**
   * タグ内テキストを取得
   *
   * このエレメントビルダで管理するタグ内のテキストノードに設定されている文字列を取得します。
   */
  getText(): string | null;

  /**
   * タグ内テキストを設定
   *
   * このエレメントビルダで管理するタグに、テキストノードを設定します。\
   * テキストノードは、子タグの末尾（末尾グループよりも後）に追加します。
   */
  setText(text: string | null): IHTMLElementBuilder;

  /**
   * タグを表示
   *
   * このエレメントビルダで管理するタグに、画面上に表示するスタイルを追加します。
   */
  show(): IHTMLElementBuilder;

  /**
   * タグを非表示
   *
   * このエレメントビルダで管理するタグに、画面上に表示しないスタイルを追加します。
   */
  hide(): IHTMLElementBuilder;

  /**
   * 子タグを先頭グループに追加
   *
   * 子タグの先頭グループに、メソッドを呼び出した順番で子タグを追加します。\
   * build メソッドで HTML タグをビルドした際、先頭グループ＋末尾グループの順で子タグを追加します。
   */
  prependChild(value: HTMLElement | IHTMLElementBuilder | IUIElement): IHTMLElementBuilder;

  /**
   * 子タグを末尾グループに追加
   *
   * 子タグの末尾グループに、メソッドを呼び出した順番で子タグを追加します。\
   * build メソッドで HTML タグをビルドした際、先頭グループ＋末尾グループの順で子タグを追加します。
   */
  appendChild(value: HTMLElement | IHTMLElementBuilder | IUIElement): IHTMLElementBuilder;

  /**
   * 子タグを削除
   *
   * prependChild, appendChild メソッドで追加された子タグ、
   * または、このエレメントビルダ初期化時にすでに追加されている子タグのうち、引数で指定されたタグを削除します。
   */
  removeChild(value: HTMLElement | IHTMLElementBuilder | IUIElement): IHTMLElementBuilder;

  /**
   * 子タグを全て削除
   *
   * prependChild, appendChild メソッドで追加された子タグ、
   * または、このエレメントビルダ初期化時にすでに追加されている子タグを全て削除します。
   */
  removeChildren(): IHTMLElementBuilder;

  /**
   * HTML タグをビルド
   *
   * 各メソッドで行われたタグに対する変更を適用します。
   */
  build(): IHTMLElementBuilder;

  /**
   * 内部情報をリセット
   *
   * 各メソッドで行われたタグに対する変更を破棄します。
   */
  reset(): IHTMLElementBuilder;
}

type HTMLElementBuilderClass = 'HTMLElementBuilder';

type IHTMLElementConfig = {
  /** クラス一覧 */
  readonly classNames: string[];

  /** タグ属性 */
  readonly attributes: {
    [key: string]: string | null;
  };

  /** 新しい値 */
  readonly newValue: string | null;

  /** 変更前のテキスト */
  readonly oldText: string | null;

  /** 変更後のテキスト */
  readonly newText: string | null;

  /** 先頭グループに追加された HTML エレメント */
  readonly reservePrepend: HTMLElement[];

  /** 末尾グループに追加された HTML エレメント */
  readonly reserveAppend: HTMLElement[];

  /** 削除された HTML エレメント */
  readonly reserveRemove: HTMLElement[];

  /** 追加された HTML エレメントビルダ */
  readonly builders: IHTMLElementBuilder[];
};

/* ********************************************************************** *
 * BrowserHelper
 * ********************************************************************** */

type BrowserInformation = {os: OperatingSystem; application: BrowserApplication};

type OperatingSystem =
  /** Unknown */
  | 'Unknown'
  /** Windows */
  | 'Windows'
  /** Mac */
  | 'Mac'
  /** iPad */
  | 'iPad'
  /** iPhone */
  | 'iPhone'
  /** Android */
  | 'Android';
type BrowserApplication =
  /** Unknown */
  | 'Unknown'
  /** IE */
  | 'MSIE'
  /** MS-Edge */
  | 'MS-Edge'
  /** Chromium-Edge */
  | 'Chromium-Edge'
  /** Firefox */
  | 'Firefox'
  /** Chrome */
  | 'Chrome'
  /** Safari */
  | 'Safari';

/* ********************************************************************** *
 * MessageBuilder
 * ********************************************************************** */

interface IMessageBuilder {
  /** メッセージを取得 */
  get(key: string, ...args: MessageBuilderArgumentType): string;

  /** メッセージを取得 */
  getLocalize(key: string, localeId: string, ...args: MessageBuilderArgumentType): string;

  /** メッセージを登録 */
  register(uuid: string): void;
}

type MessageBuilderArgumentType = (string | number | boolean)[];

/* ********************************************************************** *
 * JavaScript 互換
 * ********************************************************************** */

type CSJSPrimitiveSingleType = string | number | boolean | object | null | undefined;
type CSJSPrimitiveAllType = CSJSPrimitiveSingleType | CSJSPrimitiveSingleType[];
